import numpy as np
import pandas as pd
import warnings
from typing import Union, Optional, Callable, Tuple
from scipy.integrate import trapezoid
import matplotlib.pyplot as plt
from abc import ABC
from functools import cached_property
from lifelines.statistics import logrank_test

from SurvivalEVAL.Evaluations.custom_types import Numeric, NumericArrayLike
from SurvivalEVAL.Evaluations.util import (check_and_convert, predict_rmst, predict_mean_st, predict_median_st,
                                           predict_prob_from_curve, predict_multi_probs_from_curve,
                                           quantile_to_survival, zero_padding)

from SurvivalEVAL.Evaluations.Concordance import concordance
from SurvivalEVAL.Evaluations.AreaUnderROCurve import auc
from SurvivalEVAL.Evaluations.BrierScore import single_brier_score, brier_multiple_points
from SurvivalEVAL.Evaluations.MeanError import mean_error
from SurvivalEVAL.Evaluations.SingleTimeCalibration import one_calibration, integrated_calibration_index
from SurvivalEVAL.Evaluations.DistributionCalibration import d_calibration, km_calibration, residuals


class SurvivalEvaluator:
    def __init__(
            self,
            pred_survs: NumericArrayLike,
            time_coordinates: NumericArrayLike,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
            predict_time_method: str = "Median",
            interpolation: str = "Linear"
    ):
        """
        Initialize the Evaluator

        Parameters
        ----------
        pred_survs: NumericArrayLike,
            Accept shapes: (n_time_points,) or (n_samples, n_time_points).
            Predicted survival curves for the testing samples.
            At least one of `pred_survs` or `time_coordinates` must be a 2D array.
        time_coordinates: NumericArrayLike,
            Accept shapes: (n_time_points,) or (n_samples, n_time_points).
            Time coordinates corresponding to the survival curves.
            At least one of `pred_survs` or `time_coordinates` must be a 2D array.
        event_times: NumericArrayLike, shape = (n_samples, )
            Actual event/censor time for the testing samples.
        event_indicators: NumericArrayLike, shape = (n_samples, )
            Binary indicators of censoring for the testing samples
        train_event_times: Optional[NumericArrayLike], shape = (n_train_samples, ), default: None
            Actual event/censor time for the training samples.
        train_event_indicators: Optional[NumericArrayLike], shape = (n_train_samples, ), default: None
            Binary indicators of censoring for the training samples
        predict_time_method: str, default = "Median"
            Method for calculating predicted survival time. Available options are "Median", "Mean" and "RMST".
        interpolation: str, default = "Linear"
            Method for interpolation. Available options are ['Linear', 'Pchip'].
        """
        pred_survs = check_and_convert(pred_survs)
        time_coordinates = check_and_convert(time_coordinates)

        self.ndim_time = time_coordinates.ndim
        self.ndim_surv = pred_survs.ndim
        self._pred_survs, self._time_coordinates = zero_padding(pred_survs, time_coordinates)

        event_times, event_indicators = check_and_convert(event_times, event_indicators)
        self.event_times = event_times
        self.event_indicators = event_indicators

        if (train_event_times is not None) and (train_event_indicators is not None):
            train_event_times, train_event_indicators = check_and_convert(train_event_times, train_event_indicators)
        self.train_event_times = train_event_times
        self.train_event_indicators = train_event_indicators

        if predict_time_method == "Median":
            self.predict_time_method = predict_median_st
        elif predict_time_method == "Mean":
            self.predict_time_method = predict_mean_st
        elif predict_time_method == "RMST":
            self.predict_time_method = predict_rmst
        else:
            error = "Please enter one of 'Median', 'Mean', or 'RMST' for calculating predicted survival time."
            raise TypeError(error)

        self.interpolation = interpolation

        self._NO_CENSOR = np.all(self.event_indicators == 1)


    def _error_trainset(self, method_name: str):
        if (self.train_event_times is None) or (self.train_event_indicators is None):
            raise TypeError("Train set information is missing. "
                            "Evaluator cannot perform {} evaluation.".format(method_name))

    @property
    def pred_survs(self):
        return self._pred_survs

    @pred_survs.setter
    def pred_survs(self, val: NumericArrayLike):
        print("Setter called. Resetting predicted curves for this evaluator.")
        self._pred_survs = check_and_convert(val)
        self._clear_cache()

    @property
    def time_coordinates(self):
        return self._time_coordinates

    @time_coordinates.setter
    def time_coordinates(self, val: NumericArrayLike):
        print("Setter called. Resetting time coordinates for this evaluator.")
        self._time_coordinates = check_and_convert(val)
        self._clear_cache()

    @cached_property
    def predicted_event_times(self):
        return self.predict_time_from_curve(self.predict_time_method)

    def _clear_cache(self):
        # See how to clear cache in functools:
        # https://docs.python.org/3/library/functools.html#functools.cached_property
        # https://stackoverflow.com/questions/62662564/how-do-i-clear-the-cache-from-cached-property-decorator
        self.__dict__.pop('predicted_event_times', None)

    def predict_time_from_curve(
            self,
            predict_method: Callable,
    ) -> np.ndarray:
        """
        Predict survival time from survival curves.

        Parameters
        ----------
        predict_method: Callable
            A function that takes in a survival curve and returns a predicted survival time.
            There are three build-in methods: 'predict_median_st', 'predict_mean_st', and 'predict_rmst'.
            'predict_median_st' uses the median of the survival curve as the predicted survival time.
            'predict_mean_st' uses the expected time of the survival curve as the predicted survival time.
            'predict_rmst' uses the restricted mean survival time of the survival curve as the predicted survival time.

        Returns
        -------
        predicted_times: np.ndarray, shape = (n_samples, )
            Predicted survival time for each sample.
        """
        if ((predict_method is not predict_mean_st) and (predict_method is not predict_median_st) and
                (predict_method is not predict_rmst)):
            error = "Prediction method must be 'predict_mean_st', 'predict_median_st', or 'predict_rmst'" \
                    "got '{}' instead".format(predict_method.__name__)
            raise TypeError(error)

        predicted_times = predict_method(self._pred_survs, self._time_coordinates, self.interpolation)
        return predicted_times

    def predict_probability_from_curve(
            self,
            target_time: Union[float, int, np.ndarray],
    ) -> np.ndarray:
        """
        Calculate the survival probability at a given time point from a predicted curve.

        Each predicted curve will only have one corresponding probability.
        Note that this method is different from the 'predict_multi_probabilities_from_curve' method,
        which predicts the multiple probabilities at multiple time points from a predicted curve.

        Parameters
        ----------
        target_time: Union[float, int, np.ndarray], shape = (n_samples, )
            Time point(s) at which the probability of event is to be predicted. If float or int, the same time point is
            used for all samples. If array-like, each sample will have it own target time. The length of the array must
            be the same as the number of samples.

        Returns
        -------
        predicted_probability: np.ndarray, shape = (n_samples, )
            Predicted probabilities of event at the target time point(s).
        """
        n_samples = self._pred_survs.shape[0]
        if isinstance(target_time, (float, int)):
            target_time = target_time * np.ones(n_samples, dtype=self._time_coordinates.dtype)
        elif isinstance(target_time, np.ndarray):
            assert target_time.ndim == 1, "Target time must be a 1D array"
            assert target_time.shape[0] == self._pred_survs.shape[0], "Target time must have the same length as " \
                                                                           "the number of samples"
        else:
            error = "Target time must be a float, int, or 1D array, got '{}' instead".format(type(target_time))
            raise TypeError(error)

        predict_probs = np.empty(n_samples, dtype=self._pred_survs.dtype)

        if self.ndim_surv == 2 and self.ndim_time == 1:
            for i, curve in enumerate(self._pred_survs):
                predict_probs[i] = predict_prob_from_curve(curve, self.time_coordinates, target_time[i],
                                                           self.interpolation)
        elif self.ndim_surv == 1 and self.ndim_time == 2:
            for i, times in enumerate(self._time_coordinates):
                predict_probs[i] = predict_prob_from_curve(self._pred_survs, times, target_time[i],
                                                           self.interpolation)
        elif self.ndim_surv == 2 and self.ndim_time == 2:
            for i in range(n_samples):
                predict_probs[i] = predict_prob_from_curve(self._pred_survs[i, :], self._time_coordinates[i, :],
                                                           target_time[i], self.interpolation)
        else:
            raise TypeError("Dimensional error")

        return predict_probs

    def predict_multi_probabilities_from_curve(
            self,
            target_times: np.ndarray
    ) -> np.ndarray:
        """
        Calculate the survival probability at multiple time points from the predicted curve.

        Parameters
        ----------
        target_times: np.ndarray, shape = (n_target_times)
            Time points at which the probability of event is to be predicted.

        Returns
        -------
        prob_mat: np.ndarray, shape = (n_samples, n_target_times)
            Predicted probabilities of event at the target time points.
        """
        if self.ndim_surv == 2 and self.ndim_time == 1:
            n_samples = self._pred_survs.shape[0]
            prob_mat = np.empty((n_samples, len(target_times)), dtype=self._pred_survs.dtype)

            for i, curve in enumerate(self._pred_survs):
                prob_mat[i] = predict_multi_probs_from_curve(
                    curve, self.time_coordinates, target_times, self.interpolation
                )
        elif self.ndim_surv == 1 and self.ndim_time == 2:
            n_samples = self._time_coordinates.shape[0]
            prob_mat = np.empty((n_samples, len(target_times)), dtype=self._pred_survs.dtype)

            for i, times in enumerate(self._time_coordinates):
                prob_mat[i] = predict_multi_probs_from_curve(
                    self._pred_survs, times, target_times, self.interpolation
                )
        elif self.ndim_surv == 2 and self.ndim_time == 2:
            n_samples = self._pred_survs.shape[0]
            prob_mat = np.empty((n_samples, len(target_times)), dtype=self._pred_survs.dtype)

            for i in range(n_samples):
                prob_mat[i] = predict_multi_probs_from_curve(
                    self._pred_survs[i, :], self._time_coordinates[i, :], target_times, self.interpolation
                )
        else:
            raise TypeError("Dimensional error")

        return prob_mat

    def plot_survival_curves(
            self,
            curve_indices: Union[int, list[int]],
            color: Union[str, list[str], None] = None,
            x_lim: Union[float, tuple[float, float], None] = None,
            y_lim: Union[float, tuple[float, float], None] = None,
            x_label: str = 'Time',
            y_label: str = 'Survival probability',
            **kwargs
    ) -> (plt.Figure, plt.Axes):
        """
        Plot survival curves from the predicted survival curves.

        Parameters
        ----------
        curve_indices: Union[int, list[int]]
            Index(ces) of the curves to be plotted.
        color: Union[str, list[str], None]
            Color(s) of the curves to be plotted. If None, default color will be used.
        x_lim: Union[float, tuple[float, float], None]
            Limits for the x-axis. If None, default limits will be used.
        y_lim: Union[float, tuple[float, float], None]
            Limits for the y-axis. If None, default limits will be used.
        x_label: str
            Label for the x-axis. Default is 'Time'.
        y_label: str
            Label for the y-axis. Default is 'Survival probability'.
        kwargs: dict

        Returns
        -------
        fig, ax: tuple
            Figure and axis objects of the plot.
        """
        fig, ax = plt.subplots()
        if self.ndim_surv == 2 and self.ndim_time == 1:
            ax.plot(self._time_coordinates, self._pred_survs[curve_indices, :].T, color=color,
                    label=curve_indices, **kwargs)
        elif self.ndim_surv == 1 and self.ndim_time == 2:
            ax.plot(self._time_coordinates[curve_indices, :].T, self._pred_survs, color=color,
                    label=curve_indices, **kwargs)
        elif self.ndim_surv == 2 and self.ndim_time == 2:
            ax.plot(self._time_coordinates[curve_indices, :].T, self._pred_survs[curve_indices, :].T,
                    color=color, label=curve_indices, **kwargs)
        else:
            raise TypeError("Dimensional error")

        if y_lim is None:
            ax.set_ylim(0, 1.02)
        else:
            ax.set_ylim(y_lim)

        if x_lim is not None:
            ax.set_xlim(x_lim)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.legend()
        return fig, ax

    def plot_quantile_curves(
            self,
            curve_indices: Union[int, list[int]],
            color: Union[str, list[str], None] = None,
            x_lim: Union[float, tuple[float, float], None] = None,
            y_lim: Union[float, tuple[float, float], None] = None,
            x_label: str = 'Quantile',
            y_label: str = 'Time',
            **kwargs
    ) -> (plt.Figure, plt.Axes):
        """
        Plot quantile curves from the predicted survival curves.

        Parameters
        ----------
        curve_indices: Union[int, list[int]]
            Index(ces) of the curves to be plotted.
        color: Union[str, list[str], None]
            Color(s) of the curves to be plotted. If None, default color will be used.
        x_lim: Union[float, tuple[float, float], None]
            Limits for the x-axis. If None, default limits will be used.
        y_lim: Union[float, tuple[float, float], None]
            Limits for the y-axis. If None, default limits will be used.
        x_label: str
            Label for the x-axis. Default is 'Quantile'.
        y_label: str
            Label for the y-axis. Default is 'Time'.
        kwargs: dict

        Returns
        -------
        fig, ax: tuple
            Figure and axis objects of the plot.
        """
        pred_quan = 1 - self._pred_survs
        fig, ax = plt.subplots()
        if self.ndim_surv == 2 and self.ndim_time == 1:
            ax.plot(pred_quan[curve_indices, :].T, self._time_coordinates, color=color,
                    label=curve_indices, **kwargs)
        elif self.ndim_surv == 1 and self.ndim_time == 2:
            ax.plot(pred_quan, self._time_coordinates[curve_indices, :].T, color=color,
                    label=curve_indices, **kwargs)
        elif self.ndim_surv == 2 and self.ndim_time == 2:
            ax.plot(pred_quan[curve_indices, :].T, self._time_coordinates[curve_indices, :].T,
                    color=color, label=curve_indices, **kwargs)
        else:
            raise TypeError("Dimensional error")

        if x_lim is None:
            ax.set_xlim(0, 1)
        else:
            ax.set_xlim(x_lim)

        if y_lim is not None:
            ax.set_ylim(y_lim)
        ax.set_xlabel(x_label)
        ax.set_ylabel(y_label)
        ax.legend()
        return fig, ax

    def concordance(
            self,
            ties: str = "None",
            method: str = "Harrell"
    ) -> (float, float, int):
        """
        Calculate the concordance index between the predicted survival times and the true survival times.

        Parameters
        ----------
        ties: str, default = "None"
            A string indicating the way ties should be handled.
            Options: "None" (default), "Time", "Risk", or "All"
            "None" will throw out all ties in true survival time and all ties in predict survival times (risk scores).
            "Time" includes ties in true survival time but removes ties in predict survival times (risk scores).
            "Risk" includes ties in predict survival times (risk scores) but not in true survival time.
            "All" includes all ties.
            Note the concordance calculation is given by
            (Concordant Pairs + (Number of Ties/2))/(Concordant Pairs + Discordant Pairs + Number of Ties).
        method: str, default = "Harrell"
            A string indicating the method for constructing the pairs of samples.
            Options: "Harrell" (default) or "Margin"
            "Harrell": the pairs are constructed by comparing the predicted survival time of each sample with the
            event time of all other samples. The pairs are only constructed between samples with comparable
            event times. For example, if i-th sample has a censor time of 10, then the pairs are constructed by
            comparing the predicted survival time of sample i with the event time of all samples with event
            time of 10 or less.
            "Margin": the pairs are constructed between all samples. A best-guess time for the censored samples
            will be calculated and used to construct the pairs.
        :return: (float, float, int)
            The concordance index, the number of concordant pairs, and the number of total pairs.
        """
        # Choose prediction method based on the input argument
        # Check if there is no censored instance, if so, naive Brier score is applied
        if self._NO_CENSOR:
            method = "Harrell"

        if method == "Margin":
            self._error_trainset("margin concordance")

        return concordance(
            predicted_times=self.predicted_event_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            method=method,
            ties=ties
        )

    def auc(
            self,
            target_time: Optional[Numeric] = None
    ) -> float:
        """
        Calculate the area under the ROC curve (AUC) score at a given time point from the predicted survival curve.

        Parameters
        ----------
        target_time: float, int, or None, default = None
            Time point at which the AUC score is to be calculated. If None, the AUC score is calculated at the
            median time of all the event/censor times from the training and test sets.

        Returns
        -------
        brier_score: float
            The Brier score at the target time point.
        """
        event_times = np.concatenate((self.event_times, self.train_event_times)) \
            if self.train_event_times is not None else self.event_times

        if target_time is None:
            target_time = np.quantile(event_times, 0.5)

        predict_probs = self.predict_probability_from_curve(target_time)

        return auc(
            predict_probs=predict_probs,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            target_time=target_time
        )

    def brier_score(
            self,
            target_time: Optional[Numeric] = None,
            IPCW_weighted: bool = True
    ) -> float:
        """
        Calculate the Brier score at a given time point from the predicted survival curve.

        Parameters
        ----------
        target_time: float, int, or None, default = None
            Time point at which the Brier score is to be calculated. If None, the Brier score is calculated at the
            median time of all the event/censor times from the training and test sets.
        IPCW_weighted: bool, default = True
            Whether to use IPCW weighting for the Brier score.
        :return: float
            The Brier score at the target time point.
        """
        if self._NO_CENSOR:
            IPCW_weighted = False

        if IPCW_weighted:
            self._error_trainset("IPCW-weighted Brier score (BS)")

        if target_time is None:
            target_time = np.quantile(np.concatenate((self.event_times, self.train_event_times)), 0.5)

        predict_probs = self.predict_probability_from_curve(target_time)

        return single_brier_score(
            preds=predict_probs,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            target_time=target_time,
            ipcw=IPCW_weighted
        )

    def brier_score_multiple_points(
            self,
            target_times: np.ndarray,
            IPCW_weighted: bool = True
    ) -> np.ndarray:
        """
        Calculate multiple Brier scores at multiple specific times.

        Parameters
        ----------
        target_times: np.ndarray, default: None
            The specific time points for which to estimate the Brier scores.
        IPCW_weighted: bool, default = True
            Whether to use IPCW weighting for the Brier score.
        :return:
            Values of multiple Brier scores.
        """
        # Check if there is no censored instance, if so, naive Brier score is applied
        if self._NO_CENSOR:
            IPCW_weighted = False

        if IPCW_weighted:
            self._error_trainset("IPCW-weighted Brier score (BS)")

        pred_probs_mat = self.predict_multi_probabilities_from_curve(target_times)

        return brier_multiple_points(
            pred_mat=pred_probs_mat,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            target_times=target_times,
            ipcw=IPCW_weighted
        )

    def integrated_brier_score(
            self,
            num_points: int = None,
            IPCW_weighted: bool = True,
            draw_figure: bool = False
    ) -> float:
        """
        Calculate the integrated Brier score (IBS) from the predicted survival curve.

        Parameters
        ----------
        num_points: int, default = None
            Number of points at which the Brier score is to be calculated. If None, the number of points is set to
            the number of event/censor times from the training and test sets.
        IPCW_weighted: bool, default = True
            Whether to use IPCW weighting for the Brier score.
        draw_figure: bool, default = False
            Whether to draw the figure of the IBS.
        :return: float
            The integrated Brier score.
        """
        # Check if there is no censored instance, if so, naive Brier score is applied
        if self._NO_CENSOR:
            IPCW_weighted = False

        if IPCW_weighted:
            self._error_trainset("IPCW-weighted Integrated Brier Score (IBS)")

        max_target_time = np.max(np.concatenate((self.event_times, self.train_event_times))) if self.train_event_times \
            is not None else np.max(self.event_times)

        # If number of target time is not indicated, then we use the censored times obtained from test set
        if num_points is None:
            censored_times = self.event_times[self.event_indicators == 0]
            time_points = np.unique(censored_times)
            if time_points.size == 0:
                raise ValueError("You don't have censor data in the test set, "
                                 "please provide \"num_points\" for calculating IBS")
            else:
                time_range = np.max(time_points) - np.min(time_points)
        else:
            time_points = np.linspace(0, max_target_time, num_points)
            time_range = max_target_time

        # Get single brier score from multiple target times, and use trapezoidal integral to calculate ISB.
        #########################
        # Solution 1, implemented using metrics multiplication, this is geometrically faster than solution 2
        b_scores = self.brier_score_multiple_points(time_points, IPCW_weighted)
        if np.isnan(b_scores).any():
            warnings.warn("Time-dependent Brier Score contains nan")
            bs_dict = {}
            for time_point, b_score in zip(time_points, b_scores):
                bs_dict[time_point] = b_score
            print("Brier scores for multiple time points are:\n", bs_dict)
        integral_value = trapezoid(b_scores, time_points)
        ibs_score = integral_value / time_range
        ##########################
        # (Deprecated)
        # Solution 2, implemented by iteratively calling self.brier_score(),
        # this solution is much slower than solution 1
        # b_scores = []
        # for i in range(len(time_points)):
        #     b_score = self.brier_score(time_points[i])
        #     b_scores.append(b_score)
        # b_scores = np.array(b_scores)
        # integral_value = trapezoid(b_scores, time_points)
        # ibs_score = integral_value / time_range

        # Draw the Brier score graph
        if draw_figure:
            plt.plot(time_points, b_scores, 'bo-')
            score_text = r'IBS$= {:.3f}$'.format(ibs_score)
            plt.plot([], [], ' ', label=score_text)
            plt.legend()
            plt.xlabel('Time')
            plt.ylabel('Brier Score')
            plt.show()
        return ibs_score

    def mae(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the MAE score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the MAE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for MAE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default: False
            Whether to use log scale for the time axis.
        verbose: boolean, default: False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.            

        Returns
        -------
        mae_score: float
            The MAE score for the test set.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        if weighted is None:
            weighted = False if method == "Uncensored" or "Hinge" else True

        return mean_error(
            predicted_times=self.predicted_event_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            error_type="absolute",
            method=method,
            weighted=weighted,
            log_scale=log_scale,
            verbose=verbose,
            truncated_time=truncated_time
        )

    def mse(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the MSE score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the MSE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for MSE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default: False
            Whether to use log scale for the time axis.
        verbose: boolean, default: False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.

        Returns
        -------
        mse_score: float
            The MSE score for the test set.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        if weighted is None:
            weighted = False if method == "Uncensored" or "Hinge" else True

        return mean_error(
            predicted_times=self.predicted_event_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            error_type="squared",
            method=method,
            weighted=weighted,
            log_scale=log_scale,
            verbose=verbose,
            truncated_time=truncated_time
        )

    def rmse(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the root mean squared error (RMSE) score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the RMSE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for RMSE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default: False
            Whether to use log scale for the time axis.
        verbose: boolean, default: False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.

        Returns
        -------
        rmse_score: float
            The RMSE score for the test set.
        """
        return self.mse(method, weighted, log_scale, verbose, truncated_time) ** 0.5

    def one_calibration(
            self,
            target_time: Numeric,
            num_bins: int = 10,
            binning_strategy: str = "C",
            method: str = "DN"
    ) -> (float, list, list):
        """
        Calculate the one calibration score at a given time point from the predicted survival curve.
        Parameters
        ----------
        target_time: float, int
            Time point at which the one calibration score is to be calculated.
        num_bins: int, default: 10
            Number of bins used to calculate the one calibration score.
        binning_strategy: str
            The strategy to bin the predictions. The options are: "C" (default), and "H".
            C-statistics means the predictions are divided into equal-sized bins based on the predicted probabilities.
            H-statistics means the predictions are divided into equal-increment bins from 0 to 1.
        method: string, default: "DN"
            The method used to calculate the one calibration score.
            Options: "Uncensored", or "DN" (default)

        Returns
        -------
        p_value: float
            The p-value of the calibration test.
        observed_probabilities: list
            The observed probabilities in each bin.
        expected_probabilities: list
            The expected probabilities in each bin.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        predict_probs = self.predict_probability_from_curve(target_time)
        return one_calibration(
            preds=1 - predict_probs,
            event_time=self.event_times,
            event_indicator=self.event_indicators,
            target_time=target_time,
            num_bins=num_bins,
            binning_strategy=binning_strategy,
            method=method
        )

    def integrated_calibration_index(
            self,
            target_time: Numeric,
            knots: int = 3,
            draw_figure: Optional[bool] = True,
            figure_range: Optional[tuple] = None
    ) -> (dict, plt.figure):
        """
        Calculate the integrated one calibration index (ICI) for a given set of predictions and true event times.

        Parameters
        ----------
        target_time: Numeric
            The specific time points for which to estimate the one calibration scores.
        knots: int, default: 3
            The number of knots to use for the spline fit in the ICI calculation.
        draw_figure: bool, default: True
            Whether to draw the figure of the ICI.
        figure_range: tuple, default: None
            The range of the figure to be drawn. If None, the range will be automatically determined based on the
            predicted probabilities and the target time.

        Returns
        -------
        summary: dict
            The summary of ICI for the target time point,
            including the ICI value, the E50, E90, and the E_max values.
        fig: plt.figure
            A figure showing the graphical calibration curve.
        """
        predict_probs = self.predict_probability_from_curve(target_time)
        return integrated_calibration_index(
            preds=1 - predict_probs,
            event_time=self.event_times,
            event_indicator=self.event_indicators,
            target_time=target_time,
            knots=knots,
            draw_figure=draw_figure,
            figure_range=figure_range
        )

    def d_calibration(
            self,
            num_bins: int = 10
    ) -> (float, np.ndarray):
        """
        Calculate the D calibration score from the predicted survival curve.
        Parameters
        ----------
        num_bins: int, default: 10
            Number of bins used to calculate the D calibration score.
        Returns
        -------
        p_value: float
            The p-value of the calibration test.
        hist: np.ndarray
            The histogram of the predicted probabilities in each bin.
        """
        predict_probs = self.predict_probability_from_curve(self.event_times)
        return d_calibration(
            pred_probs=predict_probs,
            event_indicators=self.event_indicators,
            num_bins=num_bins
        )

    def residuals(
            self,
            method: str = "CoxSnell",
            draw_figure: bool = False,
    ) -> np.ndarray:
        """
        Calculate the residuals from the predicted survival (cumulative hazard) function.

        Parameters
        ----------
        method: str, default: "CoxSnell"
            The method used to calculate the residuals.
            Options: "CoxSnell" (default), "Modified CoxSnell-v1", "Modified CoxSnell-v2",
            "Martingale", "Deviance".
        draw_figure: bool, default: False
            Whether to draw the figure of the residuals.

        Returns
        -------
        residuals: np.ndarray
            The residuals calculated from the predicted survival curve.
        """
        if self._NO_CENSOR:
            method = "CoxSnell" if method in ["CoxSnell", "Modified CoxSnell-v1", "Modified CoxSnell-v2"] else method

        predict_probs = self.predict_probability_from_curve(self.event_times)
        return residuals(
            pred_probs=predict_probs,
            event_indicators=self.event_indicators,
            method=method,
            draw_figure=draw_figure
        )

    def x_calibration(
            self,
            num_bins: int = 10
    ) -> float:
        """
        Calculate the X calibration score from the predicted survival curve.
        Parameters
        ----------
        num_bins: int, default: 10
            Number of bins used to calculate the X calibration score.

        Returns
        -------
        x_cal: float
            The X calibration score, which is the sum of squared differences between the predicted and optimal
            probabilities in each bin.
        """
        _, bin_hist = self.d_calibration(num_bins)
        n_bins = bin_hist.shape[0]
        # normalize the histogram
        d_cal_pdf = bin_hist / bin_hist.sum()
        # compute the x-calibration score
        optimal = np.ones_like(d_cal_pdf) / n_bins
        x_cal = np.sum(np.square(d_cal_pdf - optimal))
        return x_cal

    def km_calibration(
            self,
            draw_figure: bool = False
    ) -> float:
        """
        Calculate the KM calibration score from the predicted survival curve.
        Parameters
        ----------
        draw_figure: bool, default: False
            Whether to draw the figure of the KM calibration curve.
        Returns
        -------
        km_cal: float
            The KM calibration score, which is the mean survival curve of the predicted survival curves.
            It is calculated by comparing the average survival curve with the Kaplan-Meier estimate of the survival
            function.
        """
        average_survival_curve = np.mean(self._pred_survs, axis=0)
        return km_calibration(
            average_survival_curve=average_survival_curve,
            time_coordinates=self.time_coordinates,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            interpolation_method=self.interpolation,
            draw_figure=draw_figure
        )

    def log_rank(
            self,
            weightings: Optional[str] = None,
            p: Optional[float] = 0,
            q: Optional[float] = 0,
    ) -> Tuple[float, float]:
        """
        Calculate the log-rank test statistic and p-value for the predicted survival curve.

        Parameters
        ----------
        weightings: str, optional
           The weighting method is for weighted log-rank test.
           Options: "None" (default), "wilcoxon", "tarone-ware", "peto", "fleming-harrington".
           None means unweighted log-rank test.
           Wilcoxon uses the number of at-risk population at each time point as the weight.
           Tarone-Ware uses the square root of the number of at-risk population at each time point as the weight.
           Peto uses the estimated survival probability as the weight.
           Fleming-Harrington uses
               w_i = S(t_i) ** p * (1 - S(t_i)) ** q
        p: float, default: 0
            The p parameter for the Fleming-Harrington weighting method.
        q: float, default: 0
            The q parameter for the Fleming-Harrington weighting method.

        Returns
        -------
        p_value: float
            The p-value of the log-rank test.
        test_statistic: float
            The test statistic of the log-rank test.
        """
        results = logrank_test(
            durations_A = self.event_times,
            durations_B = self.predicted_event_times,
            event_observed_A = self.event_indicators,
            event_observed_B = np.ones_like(self.event_indicators, dtype=bool),
            weightings=weightings,
            p=p,
            q=q
        )
        return results.p_value, results.test_statistic

class PycoxEvaluator(SurvivalEvaluator, ABC):
    def __init__(
            self,
            surv: pd.DataFrame,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
            predict_time_method: str = "Median",
            interpolation: str = "Linear"
    ):
        """
        Evaluator for survival models in PyCox packages.

        Parameters
        ----------
        surv: pd.DataFrame, shape = (n_time_points, n_samples)
            Predicted survival curves for the testing samples
            DataFrame index represents the time coordinates for the given curves.
            DataFrame value represents transpose of the survival probabilities.
        event_times: NumericArrayLike, shape = (n_samples,)
            Event times for the testing samples.
        event_indicators: NumericArrayLike, shape = (n_samples,)
            Event indicators for the testing samples.
        train_event_times: NumericArrayLike, shape = (n_samples,), optional
            Event times for the training samples.
        train_event_indicators: NumericArrayLike, shape = (n_samples,), optional
            Event indicators for the training samples.
        predict_time_method: string, default: "Median"
            The method used to calculate the predicted event time. Options: "Median" (default), "Mean", and "RMST".
        interpolation: string, default: "Linear"
            The interpolation method used to calculate the predicted event time.
            Options: "Linear" (default), "Pchip".
        """
        time_coordinates = surv.index.values
        predicted_survival_curves = surv.values.T
        # Pycox models can sometimes obtain -0 as survival probabilities. Need to convert that to 0.
        predicted_survival_curves[predicted_survival_curves < 0] = 0
        super(PycoxEvaluator, self).__init__(predicted_survival_curves, time_coordinates, event_times,
                                             event_indicators, train_event_times, train_event_indicators,
                                             predict_time_method, interpolation)


class LifelinesEvaluator(PycoxEvaluator, ABC):
    def __init__(
            self,
            surv: pd.DataFrame,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
            predict_time_method: str = "Median",
            interpolation: str = "Linear"
    ):
        """
        Evaluator for survival models in Lifelines packages.

        Parameters
        ----------
        surv: pd.DataFrame, shape = (n_time_points, n_samples)
            Predicted survival curves for the testing samples
        event_times: NumericArrayLike, shape = (n_samples,)
            Event times for the testing samples.
        event_indicators: NumericArrayLike, shape = (n_samples,)
            Event indicators for the testing samples.
        train_event_times: NumericArrayLike, shape = (n_samples,), optional
            Event times for the training samples.
        train_event_indicators: NumericArrayLike, shape = (n_samples,), optional
            Event indicators for the training samples.
        predict_time_method: string, default: "Median"
            The method used to calculate the predicted event time. Options: "Median" (default), "Mean" and "RMST".
        interpolation: string, default: "Linear"
            The interpolation method used to calculate the predicted event time.
            Options: "Linear" (default), "Pchip".
        """
        super(LifelinesEvaluator, self).__init__(surv, event_times, event_indicators, train_event_times,
                                                 train_event_indicators, predict_time_method, interpolation)


class ScikitSurvivalEvaluator(SurvivalEvaluator, ABC):
    def __init__(
            self,
            surv,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
            predict_time_method: str = "Median",
            interpolation: str = "Linear"
    ):
        """
        Evaluator for survival models in scikit-survival packages.

        Parameters
        ----------
        surv: shape = (n_samples,)
            Predicted survival curves for the testing samples from scikit-survival model.
            Each element is a scikit-survival customized object.
            '.x' attribute is the time coordinates for the given curve. '.y' attribute is the survival probabilities.
        event_times: NumericArrayLike, shape = (n_samples,)
            Event times for the testing samples.
        event_indicators: NumericArrayLike, shape = (n_samples,)
            Event indicators for the testing samples.
        train_event_times: NumericArrayLike, shape = (n_samples,), optional
            Event times for the training samples.
        train_event_indicators: NumericArrayLike, shape = (n_samples,), optional
            Event indicators for the training samples.
        predict_time_method: string, default: "Median"
            The method used to calculate the predicted event time. Options: "Median" (default), "Mean", and "RMST".
        interpolation: string, default: "Linear"
            The interpolation method used to calculate the predicted event time.
            Options: "Linear" (default), "Pchip".
        """
        time_coordinates = surv[0].x
        predict_curves = []
        for i in range(len(surv)):
            predict_curve = surv[i].y
            if False in (time_coordinates == surv[i].x):
                raise KeyError("{}-th survival curve does not have same time coordinates".format(i))
            predict_curves.append(predict_curve)
        predicted_curves = np.array(predict_curves)
        if time_coordinates[0] != 0:
            time_coordinates = np.concatenate([np.array([0]), time_coordinates], 0)
            predicted_curves = np.concatenate([np.ones([len(predicted_curves), 1]), predicted_curves], 1)
        # If some survival curves are all ones, we should do something.
        if np.any(predicted_curves[:, len(time_coordinates) - 1] == 1):
            idx_need_fix = predicted_curves[:, len(time_coordinates) - 1] == 1
            max_prob_at_end = np.max(predicted_curves[~idx_need_fix,
                                                      len(time_coordinates) - 1])
            # max_prob_at_end + (1 - max_prob_at_end) * 0.9
            predicted_curves[idx_need_fix, len(time_coordinates) - 1] = max(0.1 * max_prob_at_end + 0.9, 0.99)
        super(ScikitSurvivalEvaluator, self).__init__(predicted_curves, time_coordinates, event_times,
                                                      event_indicators, train_event_times, train_event_indicators,
                                                      predict_time_method, interpolation)


DistributionEvaluator = SurvivalEvaluator   # Alias for the SurvivalEvaluator


class PointEvaluator:
    def __init__(
            self,
            pred_times: NumericArrayLike,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
    ):
        """
        Initialize the Evaluator

        Parameters
        ----------
        pred_times: structured array, shape = (n_samples, )
            Predicted survival times for the testing samples.
        event_times: structured array, shape = (n_samples, )
            Actual event/censor time for the testing samples.
        event_indicators: structured array, shape = (n_samples, )
            Binary indicators of censoring for the testing samples
        train_event_times: structured array, shape = (n_train_samples, )
            Actual event/censor time for the training samples.
        train_event_indicators: structured array, shape = (n_train_samples, )
            Binary indicators of censoring for the training samples
        """
        self._pred_times = check_and_convert(pred_times)

        self.event_times, self.event_indicators = check_and_convert(event_times, event_indicators)

        if (train_event_times is not None) and (train_event_indicators is not None):
            train_event_times, train_event_indicators = check_and_convert(train_event_times, train_event_indicators)
        self.train_event_times = train_event_times
        self.train_event_indicators = train_event_indicators

        self._NO_CENSOR = np.all(self.event_indicators == 1)

    def _error_trainset(self, method_name: str):
        if (self.train_event_times is None) or (self.train_event_indicators is None):
            raise TypeError("Train set information is missing. "
                            "Evaluator cannot perform {} evaluation.".format(method_name))

    @property
    def pred_times(self):
        return self._pred_times

    @pred_times.setter
    def pred_times(self, pred_times):
        print("Setter called. Resetting pred_times.")
        self._pred_times = pred_times

    def concordance(
            self,
            ties: str = "None",
            method: str = "Harrell"
    ) -> (float, float, int):
        """
        Calculate the concordance index between the predicted survival times and the true survival times.

        Parameters
        ----------
        ties: str, default = "None"
            A string indicating the way ties should be handled.
            Options: "None" (default), "Time", "Risk", or "All"
            "None" will throw out all ties in true survival time and all ties in predict survival times (risk scores).
            "Time" includes ties in true survival time but removes ties in predict survival times (risk scores).
            "Risk" includes ties in predict survival times (risk scores) but not in true survival time.
            "All" includes all ties.
            Note the concordance calculation is given by
            (Concordant Pairs + (Number of Ties/2))/(Concordant Pairs + Discordant Pairs + Number of Ties).
        method: str, default = "Harrell"
            A string indicating the method for constructing the pairs of samples.
            Options: "Harrell" (default) or "Margin"
            "Harrell": the pairs are constructed by comparing the predicted survival time of each sample with the
            event time of all other samples. The pairs are only constructed between samples with comparable
            event times. For example, if i-th sample has a censor time of 10, then the pairs are constructed by
            comparing the predicted survival time of sample i with the event time of all samples with event
            time of 10 or less.
            "Margin": the pairs are constructed between all samples. A best-guess time for the censored samples
            will be calculated and used to construct the pairs.

        Returns
        -------
        concordance_index: float
            The concordance index, which is the proportion of concordant pairs among all pairs.
        concordant_pairs: float
            The number of concordant pairs.
        total_pairs: int
            The total number of comparable pairs considered in the concordance calculation.
        """
        # Check if there is no censored instance, if so, naive Brier score is applied
        if self._NO_CENSOR:
            method = "Harrell"

        if method == "Margin":
            self._error_trainset("margin concordance")

        return concordance(
            predicted_times=self._pred_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            method=method,
            ties=ties
        )

    def mae(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the MAE score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the MAE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for MAE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default: False
            Whether to use log scale for the time axis.
        verbose: boolean, default: False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.

        Returns
        -------
        mae_score: float
            The MAE score for the test set.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        if weighted is None:
            weighted = False if method == "Uncensored" or "Hinge" else True

        return mean_error(
            predicted_times=self._pred_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            error_type="absolute",
            method=method,
            weighted=weighted,
            log_scale=log_scale,
            verbose=verbose,
            truncated_time=truncated_time
        )

    def mse(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the MSE score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the MSE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for MSE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default: False
            Whether to use log scale for the time axis.
        verbose: boolean, default: False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.

        Returns
        -------
        mse_score: float
            The MSE score for the test set.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        if weighted is None:
            weighted = False if method == "Uncensored" or "Hinge" else True

        return mean_error(
            predicted_times=self._pred_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            error_type="squared",
            method=method,
            weighted=weighted,
            log_scale=log_scale,
            verbose=verbose,
            truncated_time=truncated_time
        )

    def rmse(
            self,
            method: str = "Hinge",
            weighted: bool = None,
            log_scale: bool = False,
            verbose: bool = False,
            truncated_time = None
    ) -> float:
        """
        Calculate the root mean squared error (RMSE) score for the test set.

        Parameters
        ----------
        method: string, default: "Hinge"
            The method used to calculate the MAE score.
            Options: "Uncensored", "Hinge" (default), "Margin", "IPCW-T", "IPCW-D", or "Pseudo_obs"
        weighted: bool, default: None
            Whether to use weighting scheme for MAE.
            If None, the default value is False for "Uncensored" and "Hinge" methods, and True for the rest.
        log_scale: boolean, default = False
            Whether to use log scale for the time axis.
        verbose: boolean, default = False
            Whether to show the progress bar.
        truncated_time: float, default: None
            Truncated time.

        Returns
        -------
        rmse_score: float
            The MAE score for the test set.
        """
        return self.mse(method, weighted, log_scale, verbose, truncated_time) ** 0.5

    def log_rank(
            self,
            weightings: Optional[str] = None,
            p: Optional[float] = 0,
            q: Optional[float] = 0,
    ) -> Tuple[float, float]:
        """
        Calculate the log-rank test statistic and p-value for the predicted survival curve.

        Parameters
        ----------
        weightings: str, optional
           The weighting method is for weighted log-rank test.
           Options: "None" (default), "wilcoxon", "tarone-ware", "peto", "fleming-harrington".
           None means unweighted log-rank test.
           Wilcoxon uses the number of at-risk population at each time point as the weight.
           Tarone-Ware uses the square root of the number of at-risk population at each time point as the weight.
           Peto uses the estimated survival probability as the weight.
           Fleming-Harrington uses
               w_i = S(t_i) ** p * (1 - S(t_i)) ** q
        p: float, default: 0
            The p parameter for the Fleming-Harrington weighting method.
        q: float, default: 0
            The q parameter for the Fleming-Harrington weighting method.

        Returns
        -------
        p_value: float
            The p-value of the log-rank test.
        test_statistic: float
            The test statistic of the log-rank test.
        """
        results = logrank_test(
            durations_A = self.event_times,
            durations_B = self._pred_times,
            event_observed_A = self.event_indicators,
            event_observed_B = np.ones_like(self.event_indicators, dtype=bool),
            weightings=weightings,
            p=p,
            q=q
        )
        return results.p_value, results.test_statistic


class SingleTimeEvaluator:
    def __init__(
            self,
            pred_probs: NumericArrayLike,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            target_time: Union[float, int] = None,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
    ):
        """
        Initialize the Evaluator

        Parameters
        ----------
        pred_probs: structured array, shape = (n_samples, )
            Predicted survival probability at the target time for the testing samples.
        event_times: structured array, shape = (n_samples, )
            Actual event/censor time for the testing samples.
        event_indicators: structured array, shape = (n_samples, )
            Binary indicators of censoring for the testing samples
        target_time: float, int, or None, default = None
            Time point at which the evaluation is to be performed. If None, the target time is set to the median time
        train_event_times: structured array, shape = (n_train_samples, )
            Actual event/censor time for the training samples.
        train_event_indicators: structured array, shape = (n_train_samples, )
            Binary indicators of censoring for the training samples
        """
        self._pred_probs = check_and_convert(pred_probs)
        if self._pred_probs.ndim != 1:
            raise ValueError("predicted_probs should be a 1D array-like object, "
                             "but got a {}D array-like object".format(self._pred_probs.ndim))


        self.event_times, self.event_indicators = check_and_convert(event_times, event_indicators)

        if (train_event_times is not None) and (train_event_indicators is not None):
            train_event_times, train_event_indicators = check_and_convert(train_event_times, train_event_indicators)
        self.train_event_times = train_event_times
        self.train_event_indicators = train_event_indicators

        if target_time is None:
            # set to the median time of all the event/censor times from the training and test sets
            # if train set is not provided, use test set only
            event_times = np.concatenate((self.event_times, self.train_event_times)) \
                if self.train_event_times is not None else self.event_times
            target_time = np.quantile(event_times, 0.5)
        self.target_time = target_time
        self._NO_CENSOR = np.all(self.event_indicators == 1)

    def _error_trainset(self, method_name: str):
        if (self.train_event_times is None) or (self.train_event_indicators is None):
            raise TypeError("Train set information is missing. "
                            "Evaluator cannot perform {} evaluation.".format(method_name))

    @property
    def pred_probs(self):
        return self._pred_probs

    @pred_probs.setter
    def pred_probs(self, pred_probs):
        print("Setter called. Resetting pred_probs.")
        self._pred_probs = pred_probs

    def auc(
        self,
    ) -> float:
        """
        Calculate the area under the ROC curve (AUC) score at a given time point from the predicted survival curve.

        Returns
        -------
        auc_score: float
            The AUC score at the target time point.
        """
        return auc(
            predict_probs=self._pred_probs,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            target_time=self.target_time
        )

    def brier_score(
        self,
        IPCW_weighted: bool = True
    ) -> float:
        """
        Calculate the Brier score at a given time point from the predicted survival curve.

        Parameters
        ----------
        IPCW_weighted: bool, default = True
            Whether to use IPCW weighting for the Brier score.

        Returns
        -------
        brier_score: float
            The Brier score at the target time point.
        """
        # Check if there is no censored instance, if so, naive Brier score is applied
        if self._NO_CENSOR:
            IPCW_weighted = False

        if IPCW_weighted:
            self._error_trainset("IPCW-weighted Brier score (BS)")
        return single_brier_score(
            preds=self._pred_probs,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            train_event_times=self.train_event_times,
            train_event_indicators=self.train_event_indicators,
            target_time=self.target_time,
            ipcw=IPCW_weighted)

    def one_calibration(
        self,
        num_bins: int = 10,
        binning_strategy: str = "C",
        method: str = "DN"
    ) -> (float, list, list):
        """
        Calculate the one calibration score at a given time point from the predicted survival curve.

        Parameters
        ----------
        num_bins: int, default: 10
            Number of bins used to calculate the one calibration score.
        binning_strategy: str, default: "C"
            The strategy to bin the predictions. The options are: "C" (default), and "H".
            C-statistics means the predictions are divided into equal-sized bins based on the predicted probabilities.
            H-statistics means the predictions are divided into equal-increment bins from 0 to 1.
        method: string, default: "DN"
            The method used to calculate the one calibration score.
            Options: "Uncensored", or "DN" (default)

        Returns
        -------
        p_value: float
            The p-value of the calibration test.
        observed_probabilities: list
            The observed probabilities in each bin.
        expected_probabilities: list
            The expected probabilities in each bin.
        """
        if self._NO_CENSOR:
            method = "Uncensored"

        return one_calibration(
            preds=1 - self._pred_probs,
            event_time=self.event_times,
            event_indicator=self.event_indicators,
            target_time=self.target_time,
            num_bins=num_bins,
            binning_strategy=binning_strategy,
            method=method
        )

    def integrated_calibration_index(
            self,
            knots: int = 3,
            draw_figure: Optional[bool] = True,
            figure_range: Optional[tuple] = None
    ) -> (dict, plt.figure):
        """
        Calculate the integrated one calibration index (ICI) for a given set of predictions and true event times.

        Parameters
        ----------
        knots: int, default = 3
            The number of knots to use for the spline fit. If None, the number of knots is automatically determined.
        draw_figure: bool, default = True
            Whether to create a figure showing the graphical calibration curve.
        figure_range: tuple, optional
            The range of the figure to be plotted. If None, the range is automatically determined.

        Returns
        -------
        summary: dict
            A dictionary containing the summary of ICI for the target time point, including the ICI value,
            the E50, E90, and the E_max values.
        fig: plt.figure
            A figure showing the graphical calibration curve.
        """
        return integrated_calibration_index(
            preds=1 - self._pred_probs,
            event_time=self.event_times,
            event_indicator=self.event_indicators,
            target_time=self.target_time,
            knots=knots,
            draw_figure=draw_figure,
            figure_range=figure_range
        )


class QuantileRegEvaluator(SurvivalEvaluator):
    def __init__(
            self,
            pred_regs: NumericArrayLike,
            quantile_levels: NumericArrayLike,
            event_times: NumericArrayLike,
            event_indicators: NumericArrayLike,
            train_event_times: Optional[NumericArrayLike] = None,
            train_event_indicators: Optional[NumericArrayLike] = None,
            predict_time_method: str = "Median",
            interpolation: str = "Linear"
    ):
        """
        Initialize the quantile regression evaluator.
        Parameters
        ----------
        pred_regs: structured array,
            Accept shapes: (n_quantiles,) or (n_samples, n_quantiles).
            Predicted survival curves for the testing samples.
            At least one of `pred_regs` or `quantile_levels` must be a 2D array.
        quantile_levels: structured array,
            Accept shapes: (n_quantiles,) or (n_samples, n_quantiles).
            Time coordinates corresponding to the survival curves.
            At least one of `pred_regs` or `quantile_levels` must be a 2D array.
        event_times: structured array, shape = (n_samples, )
            Actual event/censor time for the testing samples.
        event_indicators: structured array, shape = (n_samples, )
            Binary indicators of censoring for the testing samples
        train_event_times: structured array, shape = (n_train_samples, )
            Actual event/censor time for the training samples.
        train_event_indicators: structured array, shape = (n_train_samples, )
            Binary indicators of censoring for the training samples
        predict_time_method: str, default = "Median"
            Method for calculating predicted survival time. Available options are "Median", "Mean" or "RMST".
        interpolation: str, default = "Linear"
            Method for interpolation. Available options are ['Linear', 'Pchip'].
        """
        survival_level = 1 - quantile_levels
        super(QuantileRegEvaluator, self).__init__(survival_level, pred_regs, event_times,
                                                   event_indicators, train_event_times, train_event_indicators,
                                                   predict_time_method, interpolation)

    def km_calibration(
            self,
            draw_figure: bool = False
    ):
        """
        Calculate the KM calibration score from the predicted survival curve.

        Parameters
        ----------
        draw_figure: bool, default: False
            Whether to draw the figure of the KM calibration.
        Returns
        -------
        km_cal: float
            The KM calibration score, which is the mean survival curve of the predicted survival curves.
            It is calculated by comparing the average survival curve with the Kaplan-Meier estimate of the survival
            function.
        """
        unique_times = np.unique(self.event_times[self.event_indicators == 1])
        survival_curves = quantile_to_survival(1 - self._pred_survs, self.time_coordinates,
                                               unique_times, interpolate=self.interpolation)
        avg_surv = np.mean(survival_curves, axis=0)

        return km_calibration(
            average_survival_curve=avg_surv,
            time_coordinates=unique_times,
            event_times=self.event_times,
            event_indicators=self.event_indicators,
            interpolation_method=self.interpolation,
            draw_figure=draw_figure
        )



# This code is originally from paper:
# Graphical calibration curves and the integrated calibration index (ICI) for survival models
# And its original R code is in Appendix A (using splines)
# we choose the implementation using splines + CoxPH (instead of the hazard regression) because
# (1) the two methods can compariable performance and the difference is negligible (support by the paper)
# (2) as far as I know, there is no implementation of flexible adaptive hazard regression in Python
# Import necessary libraries for data handling, modeling, and plotting
import pandas as pd
import numpy as np
import lifelines
from lifelines import CoxPHFitter, WeibullAFTFitter
from patsy import dmatrix  # For spline basis matrix
import matplotlib.pyplot as plt  # For plotting
from scipy.stats import gaussian_kde  # For density plot

# ==============================================================================
# 1. Load the Data
# ==============================================================================
data = lifelines.datasets.load_gbsg2()
# preprocessing
data.rename(columns={'cens': 'event'}, inplace=True)
data['horTh'] = data['horTh'].map({'no': 0, 'yes': 1})
data['menostat'] = data['menostat'].map({'Pre': 0, 'Post': 1})
data['tgrade'] = data['tgrade'].map({'I': 1, 'II': 2, 'III': 3})
# randomly divide the data into training and validation sets
df_train = data.sample(frac=0.7, random_state=42)  # 70% for training
df_train = df_train.reset_index(drop=True)
df_test = data.drop(df_train.index)  # remaining 30% for testing
df_test = df_test.reset_index(drop=True)
x_test = df_test.drop(columns=['time', 'event']).values

# ==============================================================================
# 2. Fit a Cox Proportional Hazards Model
# ==============================================================================
cph = CoxPHFitter()
cph.fit(df_train, duration_col='time', event_col='event')

# Define time points for 1, 2, 3, 4, 5, 6 years (in days)
times = 365 * np.arange(1, 7)

# Predict the survival probability at the given time points
survs_cox = cph.predict_survival_function(x_test, times=[365]).T.values
cdfs_cox = 1 - survs_cox

# calculate the complementary log-log (CLL) transformation of the predicted probabilities
clls_cox = np.log(-np.log(survs_cox))

# ==============================================================================
# 4. Calibration: Fit Cox Models to the CLL-Transformed Predictions Using Splines
# ==============================================================================

spline_from_cox = dmatrix("bs(x, df=3, degree=3, include_intercept=False)", {"x": clls_cox}, return_type='dataframe')
design_info = spline_from_cox.design_info
df_cox = pd.concat([df_test[['time', 'event']], spline_from_cox], axis=1)
cal_cox = CoxPHFitter().fit(df_cox, duration_col='time', event_col='event')

# grid_cox = np.linspace(cdfs_cox[:, 0].quantile(0.01), cdfs_cox[:, 0].quantile(0.99), 100)
grid_cox = np.linspace(np.quantile(cdfs_cox, 0.01), np.quantile(cdfs_cox, 0.99), 100)
grid_cox_cll = np.log(-np.log(1 - grid_cox))
spline_grid_cox = dmatrix(design_info, {"x": grid_cox_cll}, return_type='dataframe')
cal_cox_pred = 1 - cal_cox.predict_survival_function(spline_grid_cox, times=[365]).T.values

# ==============================================================================
# 5. Plot Calibration Curve
# ==============================================================================
plt.figure(figsize=(8, 6))
plt.plot(grid_cox, cal_cox_pred, 'r-', label='Cox (calibrated)')
# plt.plot(grid_forest, cal_forest_pred, 'b--', label='Forest (calibrated)')
plt.plot([0,1], [0,1], 'k:', label='Ideal')  # 45-degree line: perfect calibration
plt.xlabel('Predicted probability of 1-year mortality')
plt.ylabel('Observed probability of 1-year mortality')
plt.title('1-year mortality')
plt.legend()
plt.show()

# Overlay density of Cox 1-year predicted probabilities
plt.figure()
kde = gaussian_kde(cdfs_cox[:, 0])
xvals = np.linspace(0, 1, 100)
plt.plot(xvals, kde(xvals))
plt.ylabel("Density")
plt.title("Density of Cox 1-year predicted probabilities")
plt.show()

# ==============================================================================
# 6. Calculate ICI, E50, E90 for 1-year Predictions
# ==============================================================================
# Compute calibrated predictions for the validation set
spline_val_cox = dmatrix(design_info, {"x": clls_cox}, return_type='dataframe')
# Calibrated observed probability of death within 1 year
cal_pred_cox = 1 - cal_cox.predict_survival_function(spline_val_cox, times=[365]).T.values

# Compute calibration metrics for Cox and Forest predictions
abs_err = np.abs(cdfs_cox - cal_pred_cox)
ici = abs_err.mean()
e50 = np.median(abs_err)
e90 = np.quantile(abs_err, 0.9)





#!/usr/bin/env python

import os
from setuptools import setup, find_packages
from SurvivalEVAL.version import __version__

exec(compile(open("SurvivalEVAL/version.py").read(), "SurvivalEVAL/version.py", "exec"))


def read(filename):
    with open(os.path.join(os.path.dirname(__file__), filename), encoding='utf8') as f:
        return f.read()


setup(
    name="SurvivalEVAL",
    version=__version__,
    packages=find_packages(),
    author="Shi-ang Qi, Weijie Sun",
    author_email="shiang@ualberta.ca, weijie2@ualberta.ca",
    description="The most comprehensive Python package for evaluating survival analysis models.",
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url="https://github.com/shi-ang/SurvivalEVAL",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Topic :: Scientific/Engineering",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries",
        "Topic :: Scientific/Engineering :: Information Analysis",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires='>=3.9',
    keywords="survival analysis, evaluation, metrics, survivaleval",
    license="MIT",
    install_requires=read('requirements.txt').splitlines(),
)

'''
TODO:
1. install setuptools 

python -m pip install --user --upgrade setuptools wheel

2. run the setup.py command

python setup.py sdist bdist_wheel

Test to verify it works

python -m pip install --index-url https://test.pypi.org/simple/ --no-deps Your-Package-Name

'''


from SurvivalEVAL.Evaluator import SurvivalEvaluator,  QuantileRegEvaluator
from SurvivalEVAL.Evaluator import DistributionEvaluator, PointEvaluator, SingleTimeEvaluator
from SurvivalEVAL.Evaluator import PycoxEvaluator, ScikitSurvivalEvaluator, LifelinesEvaluator
from SurvivalEVAL.IntervalCenEvaluator import IntervalCenEvaluator

from SurvivalEVAL.Evaluations.AreaUnderROCurve import auc
from SurvivalEVAL.Evaluations.BrierScore import single_brier_score, brier_multiple_points, brier_score_ic
from SurvivalEVAL.Evaluations.Concordance import concordance
from SurvivalEVAL.Evaluations.DistributionCalibration import d_calibration, km_calibration, residuals, d_cal_ic
from SurvivalEVAL.Evaluations.MeanError import mean_error
from SurvivalEVAL.Evaluations.SingleTimeCalibration import one_calibration, integrated_calibration_index, one_cal_ic

from SurvivalEVAL.version import __version__

__author__ = 'Shi-ang Qi'
__email__ = 'shiang@ualberta.ca'

__all__ = [
    'SurvivalEvaluator', 'QuantileRegEvaluator',
    'DistributionEvaluator', 'PointEvaluator', 'SingleTimeEvaluator',
    'PycoxEvaluator', 'ScikitSurvivalEvaluator', 'LifelinesEvaluator',
    'IntervalCenEvaluator',
    'auc', 'single_brier_score', 'brier_multiple_points', 'brier_score_ic',
    'concordance', 'd_calibration', 'km_calibration', 'residuals', 'd_cal_ic',
    'mean_error', 'one_calibration', 'integrated_calibration_index', 'one_cal_ic',
    '__version__'
]



#%%
import numpy as np
import matplotlib.pyplot as plt

# For preprocessing
from sklearn.preprocessing import StandardScaler
from sklearn_pandas import DataFrameMapper

import torch # For building the networks
import torchtuples as tt # Some useful functions

from pycox.datasets import metabric
from pycox.models import LogisticHazard

from SurvivalEVAL.Evaluator import PycoxEvaluator

#%%

np.random.seed(1234)
_ = torch.manual_seed(123)

#%%

df_train = metabric.read_df()
df_test = df_train.sample(frac=0.2)
df_train = df_train.drop(df_test.index)
df_val = df_train.sample(frac=0.2)
df_train = df_train.drop(df_val.index)

cols_standardize = ['x0', 'x1', 'x2', 'x3', 'x8']
cols_leave = ['x4', 'x5', 'x6', 'x7']

standardize = [([col], StandardScaler()) for col in cols_standardize]
leave = [(col, None) for col in cols_leave]

x_mapper = DataFrameMapper(standardize + leave)

x_train = x_mapper.fit_transform(df_train).astype('float32')
x_val = x_mapper.transform(df_val).astype('float32')
x_test = x_mapper.transform(df_test).astype('float32')


#%%

num_durations = 10

labtrans = LogisticHazard.label_transform(num_durations)

get_target = lambda df: (df['duration'].values, df['event'].values)
y_train = labtrans.fit_transform(*get_target(df_train))
y_val = labtrans.transform(*get_target(df_val))

train = (x_train, y_train)
val = (x_val, y_val)

# We don't need to transform the test labels
durations_test, events_test = get_target(df_test)
durations_train, events_train = get_target(df_train)

in_features = x_train.shape[1]
num_nodes = [32, 32]
out_features = labtrans.out_features
batch_norm = True
dropout = 0.1

net = tt.practical.MLPVanilla(in_features, num_nodes, out_features, batch_norm, dropout)

model = LogisticHazard(net, tt.optim.Adam(0.01), duration_index=labtrans.cuts)

batch_size = 256
epochs = 100
callbacks = [tt.cb.EarlyStopping()]

log = model.fit(x_train, y_train, batch_size, epochs, callbacks, val_data=val)

#%%

surv = model.interpolate(10).predict_surv_df(x_test)

#%%

eval = PycoxEvaluator(surv, durations_test, events_test, durations_train, events_train, predict_time_method="RMST")

#%%

cindex, concordant_pairs, total_pairs = eval.concordance(ties="None", method="Margin")
print("Concordance index is {}, meaning that the model can correctly order {} pairs among {} comparable pairs "
      "in the testing set.".format(cindex, concordant_pairs, total_pairs))

#%%

bs = eval.brier_score(target_time=175)
print("Brier score at time 175 is {}.".format(bs))

#%%

ibs = eval.integrated_brier_score(num_points=None, draw_figure=True)
print("Integrated Brier score is {}.".format(ibs))

#%%

mae = eval.mae(method='Margin')
print("MAE-margin loss is {}.".format(mae))

#%%

t_time = round(np.percentile(durations_test, 50))
p_value, ob, pre = eval.one_calibration(target_time=t_time, method="DN")
if p_value >= 0.05:
    print("The model is one-calibrated at time {}.".format(t_time))
else:
    print("The model is not one-calibrated at time {}.".format(t_time))

#%%

p_value_2, bin_statistics = eval.d_calibration()
bins = ["[0, 0.1)", "[0.1, 0.2)", "[0.2, 0.3)", "[0.3, 0.4)", "[0.4, 0.5)", "[0.5, 0.6)", "[0.6, 0.7)", "[0.7, 0.8)",
        "[0.8, 0.9)", "[0.9, 1]"]
plt.clf()
fig, ax = plt.subplots()
ax.bar(bins, bin_statistics)
plt.setp(ax.get_xticklabels(), rotation=30)
plt.ylabel("Counts in bins")
plt.show()
if p_value_2 >= 0.05:
    print("The model is d-calibrated!")
else:
    print("The model is not d-calibrated!")


from typing import List, Tuple, Union

import numpy as np
import pandas as pd
import torch

Numeric = Union[float, int, bool]
NumericArrayLike = Union[List[Numeric], Tuple[Numeric], np.ndarray, pd.Series, pd.DataFrame, torch.Tensor]



import numpy as np
import warnings
from scipy.interpolate import PchipInterpolator, interp1d

from SurvivalEVAL.Evaluations.util import interpolated_survival_curve, _check_dim_align, predict_median_st_ind

def predict_median_survival_time_new(
        survival_curves: np.ndarray,
        times_coordinates: np.ndarray,
        interpolation: str = "Linear"
) -> float:
    """
    Get the median survival time from the survival curve. The median survival time is defined as the time point where
    the survival curve crosses 0.5. The curve is first interpolated by the given monotonic cubic interpolation method
    (Linear or Pchip). Then the curve gets extroplated by the linear function of (0, 1) and the last time point. The
    median survival time is calculated by finding the time point where the survival curve crosses 0.5.
    Parameters
    ----------
    survival_curves: np.ndarray
        The survival curve of the sample. 1-D array.
    times_coordinates: np.ndarray
        The time coordinate of the survival curve. 1-D array.
    interpolation: str
        The monotonic cubic interpolation method. One of ['None', 'Linear', 'Pchip']. Default: 'Linear'.
        If 'None', use the step function to calculate the median survival time.
        If 'Linear', use the linear equation to solve the median survival time.
        If 'Pchip', use the PchipInterpolator from scipy.interpolate.
    Returns
    -------
    median_survival_time: float
        The median survival time.
    """
    _check_dim_align(survival_curves, times_coordinates)

    ndim_surv = survival_curves.ndim
    ndim_time = times_coordinates.ndim

    # If all the predicted probabilities are 1 the integral will be infinite.
    if np.all(survival_curves == 1):
        warnings.warn("All the predicted probabilities are 1, the median survival time will be infinite.")
        return np.inf

    min_prob = float(min(survival_curves))

    if 0.5 in survival_curves:
        median_probability_time = times_coordinates[np.where(survival_curves == 0.5)[0][0]]
    elif min_prob < 0.5:
        idx_before_median = np.where(survival_curves > 0.5)[0][-1]
        idx_after_median = np.where(survival_curves < 0.5)[0][0]
        min_time_before_median = times_coordinates[idx_before_median]
        max_time_after_median = times_coordinates[idx_after_median]

        if interpolation == "None":
            # find the time point where the survival curve (step function) crosses 0.5
            idx = np.where(survival_curves > 0.5)[0][0]
            median_probability_time = times_coordinates[idx]
        elif interpolation == "Linear":
            # given last time before median and first time after median, solve the linear equation
            slope = ((survival_curves[idx_after_median] - survival_curves[idx_before_median]) /
                     (max_time_after_median - min_time_before_median))
            intercept = survival_curves[idx_before_median] - slope * min_time_before_median
            median_probability_time = (0.5 - intercept) / slope
        elif interpolation == "Pchip":
            # reverse the array because the PchipInterpolator requires the x to be strictly increasing
            spline = interpolated_survival_curve(times_coordinates, survival_curves, interpolation)
            time_range = np.linspace(min_time_before_median, max_time_after_median, num=1000)
            prob_range = spline(time_range)
            inverse_spline = PchipInterpolator(prob_range[::-1], time_range[::-1])
            median_probability_time = np.array(inverse_spline(0.5)).item()
        else:
            raise ValueError("interpolation should be one of ['Linear', 'Pchip']")
    else:
        max_time = float(max(times_coordinates))
        min_prob = float(min(survival_curves))
        slope = (1 - min_prob) / (0 - max_time)
        median_probability_time = - 0.5 / slope
    return median_probability_time


def predict_median_survival_time(
        survival_curve: np.ndarray,
        times_coordinate: np.ndarray,
        interpolation: str = "Linear"
) -> float:
    """
    Get the median survival time from the survival curve. The median survival time is defined as the time point where
    the survival curve crosses 0.5. The curve is first interpolated by the given monotonic cubic interpolation method
    (Linear or Pchip). Then the curve gets extroplated by the linear function of (0, 1) and the last time point. The
    median survival time is calculated by finding the time point where the survival curve crosses 0.5.
    Parameters
    ----------
    survival_curve: np.ndarray
        The survival curve of the sample. 1-D array.
    times_coordinate: np.ndarray
        The time coordinate of the survival curve. 1-D array.
    interpolation: str
        The monotonic cubic interpolation method. One of ['Linear', 'Pchip']. Default: 'Linear'.
        If 'Linear', use the interp1d method from scipy.interpolate.
        If 'Pchip', use the PchipInterpolator from scipy.interpolate.
    Returns
    -------
    median_survival_time: float
        The median survival time.
    """
    # If all the predicted probabilities are 1 the integral will be infinite.
    if np.all(survival_curve == 1):
        warnings.warn("All the predicted probabilities are 1, the median survival time will be infinite.")
        return np.inf

    min_prob = float(min(survival_curve))

    if 0.5 in survival_curve:
        median_probability_time = times_coordinate[np.where(survival_curve == 0.5)[0][0]]
    elif min_prob < 0.5:
        idx_before_median = np.where(survival_curve > 0.5)[0][-1]
        idx_after_median = np.where(survival_curve < 0.5)[0][0]
        min_time_before_median = times_coordinate[idx_before_median]
        max_time_after_median = times_coordinate[idx_after_median]

        if interpolation == "Linear":
            # given last time before median and first time after median, solve the linear equation
            slope = ((survival_curve[idx_after_median] - survival_curve[idx_before_median]) /
                     (max_time_after_median - min_time_before_median))
            intercept = survival_curve[idx_before_median] - slope * min_time_before_median
            median_probability_time = (0.5 - intercept) / slope
        elif interpolation == "Pchip":
            # reverse the array because the PchipInterpolator requires the x to be strictly increasing
            spline = interpolated_survival_curve(times_coordinate, survival_curve, interpolation)
            time_range = np.linspace(min_time_before_median, max_time_after_median, num=1000)
            prob_range = spline(time_range)
            inverse_spline = PchipInterpolator(prob_range[::-1], time_range[::-1])
            median_probability_time = np.array(inverse_spline(0.5)).item()
        else:
            raise ValueError("interpolation should be one of ['Linear', 'Pchip']")
    else:
        max_time = float(max(times_coordinate))
        min_prob = float(min(survival_curve))
        slope = (1 - min_prob) / (0 - max_time)
        median_probability_time = - 0.5 / slope
    return median_probability_time

if __name__ == "__main__":
    # generate random survival curves and time coordinates for testing
    np.random.seed(42)
    logits = np.random.rand(10)
    pdf = logits/ logits.sum()
    survival_curves = 1 - np.cumsum(pdf)
    times_coordinates = np.linspace(0, 10, 10)
    interpolation = "Pchip"

    median_time = predict_median_survival_time(survival_curves, times_coordinates, interpolation)
    print(f"Median Survival Time: {median_time}")
    # Test the new function
    median_time_new = predict_median_survival_time_new(survival_curves, times_coordinates, interpolation)
    print(f"Median Survival Time (new): {median_time_new}")

    median_time_ind = predict_median_st_ind(survival_curves, times_coordinates, interpolation)
    print(f"Median Survival Time (index): {median_time_ind}")


from lifelines.datasets import load_rossi
import numpy as np

np.seed = 42
# Load the data and split train/test set
rossi = load_rossi()
tr, test = rossi.iloc[:300, :], rossi.iloc[300:, :]
tr_t = tr.week.values
tr_delta = tr.arrest.values
test_t = test.week.values
test_delta = test.arrest.values

from lifelines import CoxPHFitter
# Fit the model and predict the ISD curves
cph = CoxPHFitter().fit(tr, duration_col='week', event_col='arrest')
isd_curves = cph.predict_survival_function(test)

from SurvivalEVAL.Evaluations.util import predict_rmst
t_coordinate = isd_curves.index.values
survival_curves = isd_curves.values.T
t_coordinate2 = np.tile(t_coordinate, (132, 1))
a = predict_rmst(survival_curves, t_coordinate, interpolation='Pchip')
a2 = predict_rmst(survival_curves, t_coordinate2, interpolation='Pchip')

from SurvivalEVAL import LifelinesEvaluator

# Make the evaluation
evaler = LifelinesEvaluator(isd_curves, test_t, test_delta, tr_t, tr_delta, predict_time_method= "RMST")

cindex, conc_pairs, total_pairs = evaler.concordance()
print(f"C-index is {cindex}, meaning that the model can correctly order {conc_pairs} pairs among {total_pairs} comparable pairs in the testing set.")
target_t = 30
km_cal = evaler.km_calibration()
bs = evaler.brier_score(target_t)
p_value, ob, exp = evaler.one_calibration(target_t)

#%%
import numpy as np


def estimate_curve(z, e):
    '''
    Compute Kaplan-Meier curve

    Parameters
    ----------
    z: array of size N
        event time or last observation time in the dataset
    e: array of size N
        censored (0) or uncensored (1)

    Returns
    -------
    time_points: numpy.ndarray
        array of times, where each entry corresponds an uncensored event time
        empty numpy.ndarray if all data points are censored
    pi: numpy.ndarray
        array of probabilities
        empty numpy.ndarray if all data points are censored
    '''

    # inner function
    def count_num_survive(time_point, t):
        return np.count_nonzero(t >= time_point)

    # get list of uncensored event times
    time_points, counts = np.unique(z[e > 0], return_counts=True)
    if len(time_points) == 0:
        return time_points, counts

    # compute survival rate of each time by vectorize
    vfunc = np.vectorize(count_num_survive)
    vfunc.excluded.add(1)  # fix second argument
    ni = vfunc(time_points, z)
    p = 1 - counts/ni
    pi = np.cumprod(p)

    return time_points, pi

def estimate_dist(z, e, boundaries, EPS=0.0000001):
    '''
    Estimate empirical distribution by using Kaplan-Meier estimator.
    Note:
    Suppose that a patient dies at time 10
    and another patient dies at time 20.
    Let [0, 10), [10, 20), [20, 30) be time intervals.
    Then we have
      survival_rate[time 0] = 1
      survival_rate[time 10] = 0.5
      survival_rate[time 20] = 0
    Hence
      empirical distribution in time 0-10 = 0.5
      empirical distribution in time 10-20 = 0.5
      empirical distribution in time 20-30 = 0.0
    This definition is OK, because we want to train a prediction model
    such that
      survival_rate[time 0] = 1
      survival_rate[time 10] = 0.5
      survival_rate[time 20] = 0
    If we define
      empirical distribution in time 0-10 = 0.0
      empirical distribution in time 10-20 = 0.5
      empirical distribution in time 20-30 = 0.5
    then the prediction model is trained to learn that
      survival_rate[time 0] = 1
      survival_rate[time 10] = 1
      survival_rate[time 20] = 0.5
    This problem is mostly related to the uncontinous nature of
    Kaplan-Meier curve.  For uncensored time t, the Kaplan-Meier curve
    has two valid data points due to the uncontinuous curve.
    We take the lower value for such data points.

    Parameters
    ----------
    z : ndarray (float)
        Array of event times or last observation times.
        Tensor shape is [batch size].
    e : ndarray (bool)
        Array indicating censored (0) or uncensored (1).
        Tensor shape is [batch_size].
    boundaries : list (float)
        Each element in z must be STRICTLY less than boundaries[-1].

    Returns
    -------
    empirical_dist : Tensor
        Empirical distribution of size [n_bin]
    time_KM_invalid_index : int
        Fast index of the time which estimation is invalid.
        The first censored data point after the last uncensored
        data point (or the last uncensored data point if such
        data point is none) is considered last valid data point.
        empirical_dist[:time_KM_invalid_index] contains valid estimation.
        0 <= time_KM_invalid_index < n_bin by definition.
        This means that empirical_dist[-1] is always considered invalid.
    '''

    # estimate KM curve
    uncensored = e.astype(bool)
    all_censored = False
    time_points_KM, survival_rates_KM = estimate_curve(z,e)
    if len(time_points_KM) == 0:  # if all data points are censored
        time_points_KM = np.array([ z.min() ])
        all_censored = True
    if time_points_KM[0] == 0.0:
        time_points_KM[0] = EPS
    survival_rates_KM = np.append(1.0, survival_rates_KM)
    survival_rates_KM = np.append(survival_rates_KM, 0.0)

    # compute valid index
    z_censored = z[~uncensored]
    temp = z_censored[z_censored > time_points_KM[-1]]
    if all_censored or len(temp)==0:
        invalid_index = np.searchsorted(boundaries,
                                        time_points_KM[-1],
                                        side='right')
    else:
        invalid_index = np.searchsorted(boundaries,
                                        temp.min(),
                                        side='right')
    invalid_index -= 1

    # compute emprirical distribution from KM curve
    indices_tp_threshold = np.searchsorted(time_points_KM,
                                           boundaries,
                                           side='right')
    survival_rates = np.take(survival_rates_KM, indices_tp_threshold)
    survival_rates[-1] = 0.0
    empirical_dist = survival_rates[:-1] - survival_rates[1:]
    return empirical_dist, invalid_index


def KMcalibration(f_pred, z, e, boundaries, EPS=0.000001):
    '''
    Compute KM-Calibration
    Parameters
    ----------
    f_pred : ndarray
        Prediction results with n_bin+1 endpoints.
        Each row corresponds to a prediction.
        sum_i F_pred[:,i] = 1.0
        Tensor shape is [batch_size, n_bin].
    z : ndarray
        Observation time (event time or censored time)
        Tensor shape is [batch_size].
    e : ndarray
        Indicator (censored (0) or uncensored (1))
        Tensor shape is [batch_size].
    boundaries : list (float)
        Boundaries of f_pred
        Each element in z must be STRICTLY smaller than boundaries[-1]
    Returns
    -------
    KM-Calibration : Tensor
        A non-negative float number.
    '''

    e_dist, invalid_idx = estimate_dist(z, e, boundaries)
    f_pred_mean = np.mean(f_pred,0)

    # compute logarithmic loss for KM valid region
    log_empirical = np.log(e_dist[:invalid_idx]+EPS)
    log_mean_pred = np.log(f_pred_mean[:invalid_idx]+EPS)
    loss_valid = np.sum(e_dist[:invalid_idx]*(log_empirical-log_mean_pred))

    # compute logarithmic loss for KM invalid region
    sum_empirical = np.sum(e_dist[invalid_idx:])
    log_sum_empirical = np.log(sum_empirical + EPS)
    log_sum_pred = np.log(np.sum(f_pred_mean[invalid_idx:]) + EPS)
    loss_invalid = sum_empirical*(log_sum_empirical-log_sum_pred)

    return loss_valid + loss_invalid


true_times = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
true_events = np.array([1, 1, 1, 1, 1, 1, 1, 1, 1, 1])

pdf_pre = np.random.rand(len(true_times), 100)
pdfs = pdf_pre / np.sum(pdf_pre, axis=1, keepdims=True)

KMcalibration(pdfs, true_times, true_events, [11] * len(true_times))



# test interval censoring evaluation script
import numpy as np
from SurvivalEVAL import IntervalCenEvaluator
from SurvivalEVAL.Evaluations.BrierScore import brier_score_ic

# generate some random left limit and right limit times for the test data
n_samples = 100
left_limit_times = np.random.uniform(0, 10, n_samples)
right_limit_times = left_limit_times + np.random.uniform(0, 5, n_samples)
x = np.random.binomial(1, 0.5, n_samples)

train_samples = 50
tr_left_limit_times = np.random.uniform(0, 10, train_samples)
tr_right_limit_times = tr_left_limit_times + np.random.uniform(0, 5, train_samples)
x_train = np.random.binomial(1, 0.5, train_samples)
# generate some random predicted survival probabilities for the test data
# generate some random points
logits = np.random.weibull(a=1, size=2000).reshape(100, 20)
# geenrate pdf using softmax
pdf = np.exp(logits) / np.exp(logits).sum(axis=1)[:, None]
surv = 1 - np.cumsum(pdf, axis=1)
surv = surv[:, :10]

# generate some random time points
time_points = np.random.uniform(0, 6, size=1000).reshape(100, 10)
time_points = np.cumsum(time_points, axis=1)
time_points = time_points[0, :]
target_time = np.quantile(time_points, 0.5)

# create an evaluator instance
evaluator = IntervalCenEvaluator(
    pred_survs=surv,
    time_coordinates=time_points,
    left_limits=left_limit_times,
    right_limits=right_limit_times,
    train_left_limits=tr_left_limit_times,
    train_right_limits=tr_right_limit_times,
    predict_time_method="Median",
    interpolation="Linear"
)
bs = evaluator.brier_score(method="Tsouprou-marginal", x=x, x_train=x_train)
one_cal = evaluator.one_calibration(target_time=target_time)
d_cal = evaluator.d_calibration()

# bs2 = brier_score_ic(predicted_probs, left_limit_times, right_limit_times,
#                     train_left_limits=tr_left_limit_times, train_right_limits=tr_right_limit_times,
#                     x = x, x_train=x_train,
#                     method="Tsouprou-conditional")